#include <cstdlib>
#include <cmath>
#include <ctime>
#include <iostream>

#ifdef _WIN32
#include <windows.h>
#include <sstream>
#endif

#include "Clock.h"
#include "vgalib.h"

using namespace std;

int MAXX;
int MAXY;

const float velocity = 0.2;
const int maxFlakes = 262144;
const int maxCycle = 4096;
const int maxNewBurstFlakes = 10;

extern Uint8 *buffer;
extern SDL_Surface *surface;

struct debris {
    float _x;
    float _y;
    float _dx;
    float _dy;
    bool _bActive;
    int _cycles;

    debris(int x=0, int y=0) { reset(x,y); }
    void reset(int x, int y, bool active = false)
    {
	_x = x;
	_y = y;
	float vel = (rand()-RAND_MAX/2)*velocity/(RAND_MAX/2);
	float angle = 2*M_PI*rand()/RAND_MAX;
	_dx = vel*cos(angle);
	_dy = vel*sin(angle);
	_bActive = active;
	_cycles = 0;
    }
};

debris flakes[maxFlakes];

int main(int argc, char *argv[])
{
    int i, totalFlakes=0, idxFlakes=0;
    int omoux=0, omouy=0;
    int moux=0, mouy=0;
    Uint8 events=0;

    srand(time(NULL));

    switch (argc) {
    case 3:
        MAXX = atoi(argv[1]);
        MAXY = atoi(argv[2]);
	MAXX = 16*(MAXX/16);
	MAXY = 16*(MAXY/16);
        break;
    default:
        MAXX = 640;
        MAXY = 480;
        break;
    }

    // Graphics window and palette setup
    init256();
    SDL_Color colors[256];
    for (i=0; i<256; i++) {
	if (i>170) {
	    colors[i].r = 255;
	    colors[i].g = 255;
	    if (i>240)
		colors[i].b = 255;
	    else
		colors[i].b = 0;
	} else {
	    colors[i].r = (unsigned char)((i*255/170)*4/8+(255*4/8-1));
	    colors[i].g = (unsigned char)((i*255/170)*6/6+(255*0/6-1));
	    colors[i].b = 0;
	}
    }
    colors[0].r = colors[0].g = colors[0].b = 0;
    colors[255].r = colors[255].g = colors[255].b = 255;
    SDL_SetColors(surface, colors, 0, 256);

    SDL_WM_SetCaption(
	"Click anywhere with the left mouse button - ESC to exit",
	"Click anywhere with the left mouse button - ESC to exit");
    SDL_EnableUNICODE(1);

    Clock watch;
    Uint32 framesDrawn = 0;

    // Main loop
    while (1) {

	SDL_PumpEvents();

	// Abort if ESC hit
	Uint8 *keys = SDL_GetKeyState(NULL);
	if (keys[SDLK_ESCAPE])
	    break;

	// Get mouse position
	events = SDL_GetMouseState(&moux, &mouy);
	if (omoux != moux || mouy != omouy) {
	    omoux = moux;
	    omouy = mouy;
	}
#ifndef AUTO_BOOM
	// If left mouse button pressed, spawn maxNewBurstFlakes 
	if (events & SDL_BUTTON(SDL_BUTTON_LEFT)) 
#endif
	{
	    if (totalFlakes < maxFlakes-maxNewBurstFlakes) {
		for(int j=0; j<maxNewBurstFlakes; j++) {
		    totalFlakes++;
		    do {
			idxFlakes = (idxFlakes + 1) % maxFlakes;
		    } while (flakes[idxFlakes]._bActive);
		    flakes[idxFlakes].reset(moux, mouy, true);
		    //printf("Active: %d\n", totalFlakes);
		}
	    }
	}

	// Process all...
	//#pragma omp parallel for default(none) shared(flakes, MAXX, MAXY, totalFlakes)
        for (i=0; i<maxFlakes; i++) {

	    // ... active flakes:
	    if (flakes[i]._bActive) {

		// Erase old pixel, if within frame (e.g. don't clear the roof-exceeding ones)
		if (flakes[i]._x>=0. && flakes[i]._x<MAXX && flakes[i]._y>=0. && flakes[i]._y<MAXY)
		    qplot((int) flakes[i]._x, (int) flakes[i]._y, 0);

		// Update position
		flakes[i]._x += flakes[i]._dx;
		flakes[i]._y += flakes[i]._dy + flakes[i]._cycles*flakes[i]._cycles/10000000.0;

		// If this flake run for more than maxCycle, time to switch it off
		if (maxCycle == flakes[i]._cycles++) {
		    flakes[i]._bActive = false;
		    //#pragma omp atomic
		    totalFlakes--;
		    continue;
		}

		// If this flake hit the earth, time to switch it off
		if (flakes[i]._y>=MAXY) {
		    flakes[i]._bActive = false;
		    //#pragma omp atomic
		    totalFlakes--;
		    continue;
		}

		// Draw the flake, if its X-pos is within frame
		if (flakes[i]._x>=0. && flakes[i]._x<MAXX) {

		    // But only if it is "under" the roof!
		    if (flakes[i]._y>=0.) {

			// Oh, and "dim" it towards red, and eventually, black:
			int color = 255 - int(255*(flakes[i]._cycles*2)/maxCycle); 
			if (color<0) color=0;
			qplot((int) flakes[i]._x, (int) flakes[i]._y, color);
		    }
		} else {
		    // otherwise it just got outside the valid X-pos, so switch it off
		    flakes[i]._bActive = false;
		    //#pragma omp atomic
		    totalFlakes--;
		    continue;
		}
	    } 

	}
        showpage();
	framesDrawn++;
    }
#ifdef _WIN32
    stringstream speed;
    speed << "Rendering " << framesDrawn << " frames in " << (watch.readMS()/1000.0) << " seconds. (" << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
    MessageBoxA(0, (LPCSTR) speed.str().c_str(), (const char *)"Speed of rendering", MB_OK);
#else
    cout << "Rendering " << framesDrawn << " frames in ";
    cout << (watch.readMS()/1000.0) << " seconds. (";
    cout << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
#endif
    return 0;
}
